@@ -1,66 +1,20 @@
 use std::{
     convert::{TryFrom, TryInto},
-    mem,
     ops::Range,
-    sync::Arc,
 };
 
-use crate::{Arg, Bool, DecodeError, FormatSliceElement, Table};
+use crate::{Arg, DecodeError, FormatSliceElement, Table};
 use byteorder::{ReadBytesExt, LE};
 use defmt_parser::{get_max_bitfield_range, Fragment, Parameter, Type};
 
-/// List of format strings; used when decoding a `FormatSlice` (`{:[?]}`) argument
-#[derive(Debug)]
-enum FormatList<'t> {
-    /// Build the list; used when decoding the first element
-    Build { formats: Vec<&'t str> },
-    /// Use the list; used when decoding the rest of elements
-    Use {
-        formats: Vec<&'t str>,
-        cursor: usize,
-    },
-}
-
 pub(crate) struct Decoder<'t, 'b> {
     table: &'t Table,
     pub bytes: &'b [u8],
-    format_list: Option<FormatList<'t>>,
-    // below an enum tags must be included
-    below_enum: bool,
-    pub bools_tbd: Vec<Arc<Bool>>,
 }
 
-const MAX_NUM_BOOL_FLAGS: usize = 8;
-
 impl<'t, 'b> Decoder<'t, 'b> {
     pub fn new(table: &'t Table, bytes: &'b [u8]) -> Self {
-        Self {
-            table,
-            bytes,
-            format_list: None,
-            bools_tbd: Vec::new(),
-            below_enum: false,
-        }
-    }
-
-    /// Reads a byte of packed bools and unpacks them into `args` at the given indices.
-    pub fn read_and_unpack_bools(&mut self) -> Result<(), DecodeError> {
-        let bool_flags = self.bytes.read_u8()?;
-        let mut flag_index = self.bools_tbd.len();
-
-        for bool in self.bools_tbd.iter() {
-            flag_index -= 1;
-
-            // read out the leftmost unread bit and turn it into a boolean
-            let flag_mask = 1 << flag_index;
-            let nth_flag = (bool_flags & flag_mask) != 0;
-
-            bool.set(nth_flag);
-        }
-
-        self.bools_tbd.clear();
-
-        Ok(())
+        Self { table, bytes }
     }
 
     /// Sort and deduplicate `params` so that they can be interpreted correctly during decoding
@@ -73,28 +27,14 @@ impl<'t, 'b> Decoder<'t, 'b> {
         params.dedup_by(|a, b| a.index == b.index);
     }
 
-    /// Gets a format string from
-    /// - the `FormatList`, if it's in `Use` mode, or
-    /// - from `bytes` and `table` if the `FormatList` is in `Build` mode or was not provided
+    /// Gets a format string from `bytes` and `table`
     fn get_format(&mut self) -> Result<&'t str, DecodeError> {
-        if let Some(FormatList::Use { formats, cursor }) = self.format_list.as_mut() {
-            if let Some(format) = formats.get(*cursor) {
-                *cursor += 1;
-                return Ok(format);
-            }
-        }
-
-        let index = read_leb128(&mut self.bytes)?;
+        let index = self.bytes.read_u16::<LE>()? as usize;
         let format = self
             .table
             .get_without_level(index as usize)
             .map_err(|_| DecodeError::Malformed)?;
 
-        if let Some(FormatList::Build { formats }) = self.format_list.as_mut() {
-            if !self.below_enum {
-                formats.push(format)
-            }
-        }
         Ok(format)
     }
 
@@ -132,80 +72,20 @@ impl<'t, 'b> Decoder<'t, 'b> {
         &mut self,
         num_elements: usize,
     ) -> Result<Vec<FormatSliceElement<'t>>, DecodeError> {
-        if num_elements == 0 {
-            return Ok(vec![]);
-        }
-
         let format = self.get_format()?;
-
-        // let variant_format = if
         let is_enum = format.contains('|');
-        let below_enum = self.below_enum;
-
-        if is_enum {
-            self.below_enum = true;
-        }
 
         let mut elements = Vec::with_capacity(num_elements);
-        let mut formats = vec![];
-        let mut cursor = 0;
-        for i in 0..num_elements {
-            let is_first = i == 0;
-
+        for _i in 0..num_elements {
             let format = if is_enum {
                 self.get_variant(format)?
             } else {
                 format
             };
-
-            let args = if let Some(list) = &mut self.format_list {
-                match list {
-                    FormatList::Use { .. } => self.decode_format(format)?,
-
-                    FormatList::Build { formats } => {
-                        if is_first {
-                            cursor = formats.len();
-                            self.decode_format(format)?
-                        } else {
-                            let formats = formats.clone();
-                            let old = mem::replace(
-                                &mut self.format_list,
-                                Some(FormatList::Use { formats, cursor }),
-                            );
-                            let args = self.decode_format(format)?;
-                            self.format_list = old;
-                            args
-                        }
-                    }
-                }
-            } else if is_first {
-                let mut old =
-                    mem::replace(&mut self.format_list, Some(FormatList::Build { formats }));
-                let args = self.decode_format(format)?;
-                mem::swap(&mut self.format_list, &mut old);
-                formats = match old {
-                    Some(FormatList::Build { formats, .. }) => formats,
-                    _ => unreachable!(),
-                };
-                args
-            } else {
-                let formats = formats.clone();
-                let old = mem::replace(
-                    &mut self.format_list,
-                    Some(FormatList::Use { formats, cursor: 0 }),
-                );
-                let args = self.decode_format(format)?;
-                self.format_list = old;
-                args
-            };
-
+            let args = self.decode_format(format)?;
             elements.push(FormatSliceElement { format, args });
         }
 
-        if is_enum {
-            self.below_enum = below_enum;
-        }
-
         Ok(elements)
     }
 
@@ -230,35 +110,22 @@ impl<'t, 'b> Decoder<'t, 'b> {
                 Type::I32 => args.push(Arg::Ixx(self.bytes.read_i32::<LE>()? as i128)),
                 Type::I64 => args.push(Arg::Ixx(self.bytes.read_i64::<LE>()? as i128)),
                 Type::I128 => args.push(Arg::Ixx(self.bytes.read_i128::<LE>()?)),
-                // Signed isize is encoded in zigzag-encoding.
-                Type::Isize => args.push(Arg::Ixx(
-                    zigzag_decode(read_leb128(&mut self.bytes)?) as i128
-                )),
+                Type::Isize => args.push(Arg::Ixx(self.bytes.read_i32::<LE>()? as i128)),
                 Type::U8 => args.push(Arg::Uxx(self.bytes.read_u8()? as u128)),
                 Type::U16 => args.push(Arg::Uxx(self.bytes.read_u16::<LE>()? as u128)),
-                Type::U24 => {
-                    let data_low = self.bytes.read_u8()?;
-                    let data_high = self.bytes.read_u16::<LE>()?;
-                    let data = data_low as u128 | (data_high as u128) << 8;
-                    args.push(Arg::Uxx(data as u128));
-                }
                 Type::U32 => args.push(Arg::Uxx(self.bytes.read_u32::<LE>()? as u128)),
                 Type::U64 => args.push(Arg::Uxx(self.bytes.read_u64::<LE>()? as u128)),
                 Type::U128 => args.push(Arg::Uxx(self.bytes.read_u128::<LE>()? as u128)),
-                Type::Usize => args.push(Arg::Uxx(read_leb128(&mut self.bytes)? as u128)),
+                Type::Usize => args.push(Arg::Uxx(self.bytes.read_u32::<LE>()? as u128)),
                 Type::F32 => args.push(Arg::F32(f32::from_bits(self.bytes.read_u32::<LE>()?))),
                 Type::F64 => args.push(Arg::F64(f64::from_bits(self.bytes.read_u64::<LE>()?))),
-                Type::Bool => {
-                    let arc = Arc::new(Bool::FALSE);
-                    args.push(Arg::Bool(arc.clone()));
-                    self.bools_tbd.push(arc.clone());
-                    if self.bools_tbd.len() == MAX_NUM_BOOL_FLAGS {
-                        // reached end of compression block: sprinkle values into args
-                        self.read_and_unpack_bools()?;
-                    }
-                }
+                Type::Bool => args.push(Arg::Bool(match self.bytes.read_u8()? {
+                    0 => false,
+                    1 => true,
+                    _ => return Err(DecodeError::Malformed),
+                })),
                 Type::FormatSlice => {
-                    let num_elements = read_leb128(&mut self.bytes)? as usize;
+                    let num_elements = self.bytes.read_u32::<LE>()? as usize;
                     let elements = self.decode_format_slice(num_elements)?;
                     args.push(Arg::FormatSlice { elements });
                 }
@@ -268,10 +135,7 @@ impl<'t, 'b> Decoder<'t, 'b> {
                     if format.contains('|') {
                         // enum
                         let variant = self.get_variant(format)?;
-                        let below_enum = self.below_enum;
-                        self.below_enum = true;
                         let inner_args = self.decode_format(variant)?;
-                        self.below_enum = below_enum;
                         args.push(Arg::Format {
                             format: variant,
                             args: inner_args,
@@ -293,8 +157,7 @@ impl<'t, 'b> Decoder<'t, 'b> {
                     data = match size_after_truncation {
                         1 => self.bytes.read_u8()? as u128,
                         2 => self.bytes.read_u16::<LE>()? as u128,
-                        3 => self.bytes.read_u24::<LE>()? as u128,
-                        4 => self.bytes.read_u32::<LE>()? as u128,
+                        3..=4 => self.bytes.read_u32::<LE>()? as u128,
                         5..=8 => self.bytes.read_u64::<LE>()? as u128,
                         9..=16 => self.bytes.read_u128::<LE>()? as u128,
                         _ => unreachable!(),
@@ -305,7 +168,7 @@ impl<'t, 'b> Decoder<'t, 'b> {
                     args.push(Arg::Uxx(data));
                 }
                 Type::Str => {
-                    let str_len = read_leb128(&mut self.bytes)? as usize;
+                    let str_len = self.bytes.read_u32::<LE>()? as usize;
                     let mut arg_str_bytes = vec![];
 
                     // note: went for the suboptimal but simple solution; optimize if necessary
@@ -320,7 +183,7 @@ impl<'t, 'b> Decoder<'t, 'b> {
                     args.push(Arg::Str(arg_str));
                 }
                 Type::IStr => {
-                    let str_index = read_leb128(&mut self.bytes)? as usize;
+                    let str_index = self.bytes.read_u16::<LE>()? as usize;
 
                     let string = self
                         .table
@@ -331,7 +194,7 @@ impl<'t, 'b> Decoder<'t, 'b> {
                 }
                 Type::U8Slice => {
                     // only supports byte slices
-                    let num_elements = read_leb128(&mut self.bytes)? as usize;
+                    let num_elements = self.bytes.read_u32::<LE>()? as usize;
                     let mut arg_slice = vec![];
 
                     // note: went for the suboptimal but simple solution; optimize if necessary
@@ -371,6 +234,27 @@ impl<'t, 'b> Decoder<'t, 'b> {
 
                     args.push(Arg::Preformatted(data.into()));
                 }
+                Type::FormatSequence => {
+                    let mut seq_args = Vec::new();
+                    loop {
+                        let index = self.bytes.read_u16::<LE>()? as usize;
+                        if index == 0 {
+                            break;
+                        }
+
+                        let format = self
+                            .table
+                            .get_without_level(index as usize)
+                            .map_err(|_| DecodeError::Malformed)?;
+
+                        let inner_args = self.decode_format(format)?;
+                        seq_args.push(Arg::Format {
+                            format,
+                            args: inner_args,
+                        });
+                    }
+                    args.push(Arg::FormatSequence { args: seq_args })
+                }
             }
         }
 
@@ -434,18 +318,6 @@ fn merge_bitfields(params: &mut Vec<Parameter>) {
     params.append(&mut merged_bitfields);
 }
 
-pub fn read_leb128(bytes: &mut &[u8]) -> Result<u64, DecodeError> {
-    match leb128::read::unsigned(bytes) {
-        Ok(val) => Ok(val),
-        Err(leb128::read::Error::Overflow) => Err(DecodeError::Malformed),
-        Err(leb128::read::Error::IoError(io)) => Err(io.into()),
-    }
-}
-
-fn zigzag_decode(unsigned: u64) -> i64 {
-    (unsigned >> 1) as i64 ^ -((unsigned & 1) as i64)
-}
-
 #[cfg(test)]
 mod tests {
     use super::*;
